package positive;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;

import positive.Stuff.PEnemyBullet;
import positive.Stuff.PInfo;
import positive.Stuff.Utils;

import robocode.AdvancedRobot;
import robocode.HitRobotEvent;
import robocode.HitWallEvent;
import robocode.ScannedRobotEvent;

/*
 * Enterprise
 * Original version by Positive
 *
 * Feel free (actually, I'd be pleased if you did) to modify this
 * code and post new versions of it. :)
 *
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

public class Enterprise extends AdvancedRobot
{
	private double radarTurnWanted = 10000;
	private double lastEnergyOpponent = 100;
	private double battleFieldHeight;
	private double battleFieldWidth;
	private Point2D.Double lastEnemyPosition;
	private ArrayList<PEnemyBullet> bulletsToNeutralize =
		new ArrayList<PEnemyBullet>();

	public void run()
	{
		setColors(Color.GREEN.brighter(), // bodyColor
				Color.blue, // gunColor
				Color.cyan, // radarColor
				Color.white, // bulletColor
				Color.LIGHT_GRAY) ; // scanArcColor

        setAdjustGunForRobotTurn(true);
        setAdjustRadarForGunTurn(true);

        battleFieldHeight = getBattleFieldHeight();
        battleFieldWidth = getBattleFieldWidth();

		while(true)
		{
			PInfo myPhysical = getMyPhysical();
			PInfo nextPhysical = getNextInfo(myPhysical);

			setMaxVelocity(nextPhysical.velocity);
			setAhead(1000);

			setTurnRightRadians(Utils.HeadHeadToRotation(
					getHeadingRadians(),nextPhysical.heading));
			setTurnRadarRightRadians(radarTurnWanted);

			double wantedGunHeading = -1;
			if(lastEnemyPosition!=null)
			{
				wantedGunHeading =
					Utils.pointPointToHeading(myPhysical.position,
					lastEnemyPosition);

				if(bulletsToNeutralize.size()!=0)
				{
					PEnemyBullet bullet = bulletsToNeutralize.get(0);

					double nowNeededGunHeading =
						getNeutralizeHeading(myPhysical.position,bullet,getTime());

					if(Math.abs(
							getGunHeadingRadians()-nowNeededGunHeading)<0.0001)
						// roughly the same
					{
						setFireBullet(0.1);
						bulletsToNeutralize.remove(0);
					}
				}

				if(bulletsToNeutralize.size()!=0)
				{
					PEnemyBullet nextBullet = bulletsToNeutralize.get(0);
					wantedGunHeading =
						getNeutralizeHeading(nextPhysical.position,nextBullet,getTime()+1);
				}

				setTurnGunRightRadians(
						Utils.HeadHeadToRotation(getGunHeadingRadians(),
								wantedGunHeading));

			}

			execute();
		}
	}
	public void onScannedRobot(ScannedRobotEvent e)
	{
        double headingToOpp = Utils.NormalizeHeading(
        	e.getBearingRadians() + getHeadingRadians());
        radarTurnWanted =
        	Utils.HeadHeadToRotation(getRadarHeadingRadians(),
        			headingToOpp);
        radarTurnWanted+=Utils.sign(radarTurnWanted)*Math.PI/18;
        	// proven to never lose track of opponent :)

        Point2D.Double myPosition = new Point2D.Double(getX(),getY());

        double bulletPower = lastEnergyOpponent - e.getEnergy();

        if(bulletPower>=0.1 && bulletPower<=3.0 && lastEnemyPosition!=null)
        {
        	double bulletVelocity = 20 - bulletPower*3;
        	double bulletDist = bulletVelocity + 18;
        	PInfo myTestInfo = getMyPhysical();
        	while(true)
        	{
        		double distSquared =
        			Utils.square(lastEnemyPosition.x - myTestInfo.position.x) +
        			Utils.square(lastEnemyPosition.y - myTestInfo.position.y);
        		if(distSquared<bulletDist*bulletDist)
        			break;
        		bulletDist+=bulletVelocity;
        		myTestInfo = getNextInfo(myTestInfo);
        	}
        	double bulletHeading = Utils.pointPointToHeading(
        			lastEnemyPosition,myTestInfo.position);
        	PEnemyBullet bullet = new PEnemyBullet(lastEnemyPosition,
        			e.getTime()-1,
        			bulletHeading,
        			bulletVelocity);
        	bulletsToNeutralize.add(bullet);
        }

        lastEnemyPosition =
        	Utils.PosDistHeadToPosition(myPosition,e.getDistance(),headingToOpp);

        lastEnergyOpponent = e.getEnergy();
	}
	private double getNeutralizeHeading(
			Point2D.Double fireLocation,
			PEnemyBullet enemyBullet,
			long fireTurn)
	{

		double myBulletDist = 11;
		double myBulletVelocity = 20 - 0.1 * 3;
		Point2D.Double enemyBulletPosition;
		for(double t=fireTurn-0.5;true;t++)
		{
			enemyBulletPosition = enemyBullet.getPosition(t);
			if((Utils.square(enemyBulletPosition.x - fireLocation.x)+
					Utils.square(enemyBulletPosition.y - fireLocation.y))
					< myBulletDist*myBulletDist)
				break;
			myBulletDist += myBulletVelocity;
		}

		return Utils.pointPointToHeading(fireLocation, enemyBulletPosition);
	}
	public PInfo getNextInfo(PInfo pInfo)
	{
		double wantedHeading;	// 180 degrees

		if(pInfo.position.y==18)
		{
			if(pInfo.position.x!=18)
				wantedHeading = Math.PI*1.5;
			else
				wantedHeading = 0;
		}
		else if(pInfo.position.x==18)
		{
			if(pInfo.position.y!=(battleFieldHeight-18))
				wantedHeading = 0;
			else
				wantedHeading = Math.PI/2;
		}
		else if(pInfo.position.y==(battleFieldHeight-18))
		{
			if(pInfo.position.x!=(battleFieldWidth-18))
				wantedHeading = Math.PI/2;
			else
				wantedHeading = Math.PI;
		}
		else
		{
			wantedHeading = Math.PI;
		}

		double wantedRotation =
			Utils.HeadHeadToRotation(pInfo.heading, wantedHeading);
		double maxRotation = Utils.vel2MaxRot(pInfo.velocity);
		double nextRotation;
		if(wantedRotation>maxRotation)
			nextRotation = maxRotation;
		else if(wantedRotation<-maxRotation)
			nextRotation = -maxRotation;
		else
			nextRotation = wantedRotation;

		double nextHeading =
			Utils.NormalizeHeading(pInfo.heading+nextRotation);

		double wantedDistance = 0;
		if(nextHeading==0)						// 0 degrees
			wantedDistance = (battleFieldHeight-18)-pInfo.position.y;
		else if(nextHeading==Math.PI/2)			// 90 degrees
			wantedDistance = (battleFieldWidth-18)-pInfo.position.x;
		else if(nextHeading==Math.PI)			// 180 degrees
			wantedDistance = (pInfo.position.y-18);
		else if(nextHeading==Math.PI+Math.PI/2) // 270 degrees
			wantedDistance = (pInfo.position.x-18);

		double nextVelocity =
			Utils.getNewVelocity(pInfo.velocity, wantedDistance);

		return new PInfo(Utils.PosDistHeadToPosition(
				pInfo.position,nextVelocity,nextHeading),
				nextHeading,nextVelocity);
	}
	private PInfo getMyPhysical()
	{
		return new PInfo(new Point2D.Double(getX(),getY()),
				getHeadingRadians(),
				getVelocity());
	}
}
