package positive.Stuff;

import java.awt.geom.Point2D;

import positive.General.Position2D;
import robocode.Rules;

public class Utils {

	static final public Point2D.Double PosDistHeadToPosition(
			final Point2D.Double sourcelocation, 
			final double distance, 
			final double headingRadians)
	{
		return new Point2D.Double(
				sourcelocation.x + distance * Math.sin(headingRadians),
				sourcelocation.y + distance * Math.cos(headingRadians));
	}

	static final public double HeadHeadToRotation(
			double sourceheadingRadians, 
			double destinationheadingRadians)
	{
		double rotation;
		if(sourceheadingRadians < destinationheadingRadians)
		{
			rotation = destinationheadingRadians-sourceheadingRadians;
			if(rotation < Math.PI)
				return rotation;
			else
				return rotation-Math.PI*2;
		}
		else
		{
			rotation = sourceheadingRadians-destinationheadingRadians;
			if(rotation < Math.PI)
				return -rotation;
			else
				return Math.PI*2-rotation;
		}
	}
	static final public double sign(double number)
	{
		if(number<0) return -1;
		else return 1;
	}
	static public double NormalizeHeading(double heading)
	{
		double normalizedHeading = heading;
		if(normalizedHeading < 0)
		{
			do
				normalizedHeading+=Math.PI*2;
			while(normalizedHeading < 0);
		}
		else
		{
			while(normalizedHeading >= Math.PI*2)
				normalizedHeading -= Math.PI*2;
		}
		return normalizedHeading;
	}
    public static double getNewVelocity(double velocity, double distance) {
        
        if (distance < 0) 
           return -getNewVelocity(-velocity, -distance);
     	// If the distance is negative, then change it to be positive
     	// and change the sign of the input velocity and the result
        
        if(distance<=2.0)
        	return distance;
     
        final double goalVel;
        if(distance == Double.POSITIVE_INFINITY)
           goalVel = 8.0;
        else 
           goalVel = Math.min(getMaxVelocity(distance),
            	8.0);
        
        if(velocity >= 0)
           return Math.max(velocity - Rules.DECELERATION,
              Math.min(goalVel, velocity + Rules.ACCELERATION));
     //else
        return Math.max(velocity - Rules.ACCELERATION,
                 Math.min(goalVel, velocity +  Rules.DECELERATION));
     }
    private  final static double getMaxVelocity(double distance)
     {
        final double decelTime =  Math.max(1,Math.ceil(
           //sum of 0... decelTime, solving for decelTime using quadratic formula
           (Math.sqrt((4*2/Rules.DECELERATION)*distance + 1) - 1)/2));
           
        final double decelDist = (decelTime / 2.0) * (decelTime-1) // sum of 0..(decelTime-1)
           * Rules.DECELERATION;
       
        return ((decelTime - 1) * Rules.DECELERATION) +
           ((distance - decelDist) / decelTime);
     }
    public final static double vel2MaxRot(double velocity)
    {
    	return (10-0.75*Math.abs(velocity))/180*Math.PI;
    }
    public final static double square(double number)
    {
    	return number*number;
    }
    public final static double pointPointToHeading(Point2D.Double source,Point2D.Double dest)
    {
    	return NormalizeHeading(Math.atan2(dest.x - source.x, 
        				dest.y - source.y));
    }


}
